@@ -43,8 +43,18 @@
  * which puts the surface into the 'free' state.
  *
  */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include "qxl.h"
 
+#ifdef DEBUG_SURFACE_LIFECYCLE
+#include <stdio.h>
+
+static FILE* surface_log;
+#endif
+
 typedef struct evacuated_surface_t evacuated_surface_t;
 
 struct qxl_surface_t
@@ -72,11 +82,24 @@ struct qxl_surface_t
     int			ref_count;
 
     PixmapPtr		pixmap;
-    
+
+    struct evacuated_surface_t *evacuated;
+
     union
     {
 	qxl_surface_t *copy_src;
 	Pixel	       solid_pixel;
+
+	struct
+	{
+	    int			op;
+	    PicturePtr		src_picture;
+	    PicturePtr		mask_picture;
+	    PicturePtr		dest_picture;
+	    qxl_surface_t	*src;
+	    qxl_surface_t	*mask;
+	    qxl_surface_t	*dest;
+	} composite;
     } u;
 };
 
@@ -86,6 +109,7 @@ struct evacuated_surface_t
     PixmapPtr		 pixmap;
     int			 bpp;
 
+    evacuated_surface_t *prev;
     evacuated_surface_t *next;
 };
 
@@ -115,15 +139,59 @@ struct surface_cache_t
     qxl_surface_t *cached_surfaces[N_CACHED_SURFACES];
 };
 
+#ifdef DEBUG_SURFACE_LIFECYCLE
+static void debug_surface_open(void)
+{
+    if (surface_log)
+        return;
+    surface_log = fopen("/tmp/xf86-video-qxl.surface.log", "w+");
+    if (!surface_log)
+    {
+        fprintf(stderr, "error creating surface log file (DEBUG_SURFACE_LIFECYCLE)\n");
+        exit(-1);
+    }
+}
+
+static int surface_count(qxl_surface_t *surface)
+{
+    int i;
+
+    for (i = 0; surface ;++i, surface = surface->next);
+    return i;
+}
+
+static void debug_surface_log(surface_cache_t *cache)
+{
+    int  live_n, free_n;
+
+    debug_surface_open();
+    live_n = surface_count(cache->live_surfaces);
+    free_n = surface_count(cache->free_surfaces);
+    fprintf(surface_log, "live,free,sum = %d, %d, %d\n", live_n, free_n,
+            live_n + free_n);
+    fflush(surface_log);
+}
+
+#else
+#define debug_surface_log(cache)
+#endif
+
+
 static Bool
 surface_cache_init (surface_cache_t *cache, qxl_screen_t *qxl)
 {
     int n_surfaces = qxl->rom->n_surfaces;
     int i;
 
-    cache->all_surfaces = calloc (n_surfaces, sizeof (qxl_surface_t));
     if (!cache->all_surfaces)
-	return FALSE;
+    {
+        /* all_surfaces is not freed when evacuating, since surfaces are still
+         * tied to pixmaps that may be destroyed after evacuation before
+         * recreation */
+        cache->all_surfaces = calloc (n_surfaces, sizeof (qxl_surface_t));
+        if (!cache->all_surfaces)
+            return FALSE;
+    }
 
     memset (cache->all_surfaces, 0, n_surfaces * sizeof (qxl_surface_t));
     memset (cache->cached_surfaces, 0, N_CACHED_SURFACES * sizeof (qxl_surface_t *));
@@ -137,6 +205,7 @@ surface_cache_init (surface_cache_t *cache, qxl_screen_t *qxl)
 	cache->all_surfaces[i].cache = cache;
 	cache->all_surfaces[i].dev_image = NULL;
 	cache->all_surfaces[i].host_image = NULL;
+	cache->all_surfaces[i].evacuated = NULL;
 	
 	REGION_INIT (
 	    NULL, &(cache->all_surfaces[i].access_region), (BoxPtr)NULL, 0);
@@ -161,6 +230,7 @@ qxl_surface_cache_create (qxl_screen_t *qxl)
     if (!cache)
 	return NULL;
 
+    memset(cache, 0, sizeof(*cache));
     cache->qxl = qxl;
     if (!surface_cache_init (cache, qxl))
     {
@@ -243,7 +313,7 @@ get_formats (int bpp, SpiceBitmapFmt *format, pixman_format_code_t *pformat)
 	break;
     }
 }
-		 
+
 static qxl_surface_t *
 surface_get_from_cache (surface_cache_t *cache, int width, int height, int bpp)
 {
@@ -278,12 +348,18 @@ qxl_surface_recycle (surface_cache_t *cache, uint32_t id)
     qxl_surface_t *surface = cache->all_surfaces + id;
 
     n_live--;
-    qxl_free (cache->qxl->surf_mem, surface->address);
+    qxl_free (cache->qxl->surf_mem, surface->address, "surface memory");
 
     surface->next = cache->free_surfaces;
     cache->free_surfaces = surface;
 }
 
+/*
+ * mode is used for the whole virtual screen, not for a specific head.
+ * For a single head where virtual size is equal to the head size, they are
+ * equal. For multiple heads this mode will not match any existing heads and
+ * will be the containing virtual size.
+ */
 qxl_surface_t *
 qxl_surface_cache_create_primary (surface_cache_t	*cache,
 				  struct QXLMode	*mode)
@@ -321,16 +397,31 @@ qxl_surface_cache_create_primary (surface_cache_t	*cache,
     create->type = QXL_SURF_TYPE_PRIMARY;
     create->mem = physical_address (cache->qxl, cache->qxl->ram, cache->qxl->main_mem_slot);
 
-    qxl_create_primary(qxl);
+    qxl_io_create_primary(qxl);
 
     dev_addr = (uint8_t *)qxl->ram + mode->stride * (mode->y_res - 1);
 
     dev_image = pixman_image_create_bits (format, mode->x_res, mode->y_res,
 					  (uint32_t *)dev_addr, -mode->stride);
 
+    if (qxl->fb != NULL)
+        free(qxl->fb);
+    qxl->fb = calloc (qxl->virtual_x * qxl->virtual_y, 4);
+    if (!qxl->fb)
+        return NULL;
+
     host_image = pixman_image_create_bits (format, 
 					   qxl->virtual_x, qxl->virtual_y,
-					   qxl->fb, qxl->stride);
+					   qxl->fb, mode->stride);
+#if 0
+    xf86DrvMsg(cache->qxl->pScrn->scrnIndex, X_ERROR,
+               "testing dev_image memory (%d x %d)\n",
+               mode->x_res, mode->y_res);
+    memset(qxl->ram, 0, mode->stride * mode->y_res);
+    xf86DrvMsg(cache->qxl->pScrn->scrnIndex, X_ERROR,
+               "testing host_image memory\n");
+    memset(qxl->fb, 0, mode->stride * mode->y_res);
+#endif
 
     surface = malloc (sizeof *surface);
     surface->id = 0;
@@ -340,6 +431,7 @@ qxl_surface_cache_create_primary (surface_cache_t	*cache,
     surface->bpp = mode->bits;
     surface->next = NULL;
     surface->prev = NULL;
+    surface->evacuated = NULL;
     
     REGION_INIT (NULL, &(surface->access_region), (BoxPtr)NULL, 0);
     surface->access_type = UXA_ACCESS_RO;
@@ -353,9 +445,7 @@ make_surface_cmd (surface_cache_t *cache, uint32_t id, QXLSurfaceCmdType type)
     struct QXLSurfaceCmd *cmd;
     qxl_screen_t *qxl = cache->qxl;
 
-    qxl_garbage_collect (qxl);
-    
-    cmd = qxl_allocnf (qxl, sizeof *cmd);
+    cmd = qxl_allocnf (qxl, sizeof *cmd, "surface command");
 
     cmd->release_info.id = pointer_to_u64 (cmd) | 2;
     cmd->type = type;
@@ -400,7 +490,8 @@ make_drawable (qxl_screen_t *qxl, int surface, uint8_t type,
     struct QXLDrawable *drawable;
     int i;
     
-    drawable = qxl_allocnf (qxl, sizeof *drawable);
+    drawable = qxl_allocnf (qxl, sizeof *drawable, "drawable command");
+    assert(drawable);
     
     drawable->release_info.id = pointer_to_u64 (drawable);
     
@@ -538,9 +629,9 @@ surface_send_create (surface_cache_t *cache,
     /* the final + stride is to work around a bug where the device apparently 
      * scribbles after the end of the image
      */
-    qxl_garbage_collect (cache->qxl);
+    qxl_garbage_collect (qxl);
 retry2:
-    address = qxl_alloc (qxl->surf_mem, stride * height + stride);
+    address = qxl_alloc (qxl->surf_mem, stride * height + stride, "surface memory");
 
     if (!address)
     {
@@ -549,7 +640,7 @@ retry2:
 	if (qxl_garbage_collect (qxl))
 	    goto retry2;
 
-	ErrorF ("- OOM at %d %d %d\n", width, height, bpp);
+	ErrorF ("- OOM at %d %d %d (= %d bytes)\n", width, height, bpp, width * height * (bpp / 8));
 	print_cache_info (cache);
 	
 	if (qxl_handle_oom (qxl))
@@ -572,7 +663,7 @@ retry:
 	if (!qxl_handle_oom (cache->qxl))
 	{
 	    ErrorF ("  Out of surfaces\n");
-	    qxl_free (qxl->surf_mem, address);
+	    qxl_free (qxl->surf_mem, address, "surface memory");
 	    return NULL;
 	}
 	else
@@ -594,7 +685,8 @@ retry:
 
     push_surface_cmd (cache, cmd);
 
-    dev_addr = (uint32_t *)((uint8_t *)surface->address + stride * (height - 1));
+    dev_addr
+	= (uint32_t *)((uint8_t *)surface->address + stride * (height - 1));
 
     surface->dev_image = pixman_image_create_bits (
 	pformat, width, height, dev_addr, - stride);
@@ -622,10 +714,11 @@ qxl_surface_create (surface_cache_t *    cache,
     
     if ((bpp & 3) != 0)
     {
-	ErrorF ("   Bad bpp: %d (%d)\n", bpp, bpp & 7);
+	ErrorF ("%s: Bad bpp: %d (%d)\n", __FUNCTION__, bpp, bpp & 7);
 	return NULL;
     }
 
+#if 0
     if (bpp == 8)
       {
 	static int warned;
@@ -637,23 +730,24 @@ qxl_surface_create (surface_cache_t *    cache,
 	
 	return NULL;
       }
+#endif
     
     if (bpp != 8 && bpp != 16 && bpp != 32 && bpp != 24)
     {
-	ErrorF ("   Unknown bpp\n");
+	ErrorF ("%s: Unknown bpp\n", __FUNCTION__);
 	return NULL;
     }
 
     if (width == 0 || height == 0)
     {
-	ErrorF ("   Zero width or height\n");
+	ErrorF ("%s: Zero width or height\n", __FUNCTION__);
 	return NULL;
     }
 
     if (!(surface = surface_get_from_cache (cache, width, height, bpp)))
 	if (!(surface = surface_send_create (cache, width, height, bpp)))
 	    return NULL;
-    
+
     surface->next = cache->live_surfaces;
     surface->prev = NULL;
     if (cache->live_surfaces)
@@ -674,10 +768,15 @@ qxl_surface_set_pixmap (qxl_surface_t *surface, PixmapPtr pixmap)
 static void
 unlink_surface (qxl_surface_t *surface)
 {
-    if (surface->prev)
-	surface->prev->next = surface->next;
-    else
-	surface->cache->live_surfaces = surface->next;
+    if (surface->id != 0)
+    {
+        if (surface->prev)
+            surface->prev->next = surface->next;
+        else
+            surface->cache->live_surfaces = surface->next;
+    }
+
+    debug_surface_log(surface->cache);
     
     if (surface->next)
 	surface->next->prev = surface->prev;
@@ -689,7 +788,7 @@ unlink_surface (qxl_surface_t *surface)
 }
 
 static void
-send_destroy (qxl_surface_t *surface)
+surface_destroy (qxl_surface_t *surface)
 {
     struct QXLSurfaceCmd *cmd;
 
@@ -697,9 +796,12 @@ send_destroy (qxl_surface_t *surface)
 	pixman_image_unref (surface->dev_image);
     if (surface->host_image)
 	pixman_image_unref (surface->host_image);
-    
+
+#if 0
+    ErrorF("destroy %ld\n", (long int)surface->end - (long int)surface->address);
+#endif
     cmd = make_surface_cmd (surface->cache, surface->id, QXL_SURFACE_CMD_DESTROY);
-    
+
     push_surface_cmd (surface->cache, cmd);
 }
 
@@ -726,7 +828,6 @@ surface_add_to_cache (qxl_surface_t *surface)
     
     if (n_surfaces == N_CACHED_SURFACES)
     {
-	int i;
 	destroy_id = cache->cached_surfaces[oldest]->id;
 	
 	destroy_surface = cache->cached_surfaces[oldest];
@@ -782,16 +883,39 @@ qxl_surface_unref (surface_cache_t *cache, uint32_t id)
 	qxl_surface_t *surface = cache->all_surfaces + id;
 
 	if (--surface->ref_count == 0)
-	    send_destroy (surface);
+	    surface_destroy (surface);
     }
 }
 
 void
 qxl_surface_kill (qxl_surface_t *surface)
 {
+    struct evacuated_surface_t *ev = surface->evacuated;
+
+    if (ev)
+    {
+        /* server side surface is already destroyed (via reset), don't
+         * resend a destroy. Just mark surface as not to be recreated */
+        ev->pixmap = NULL;
+        if (ev->image)
+            pixman_image_unref (ev->image);
+        if (ev->next)
+            ev->next->prev = ev->prev;
+        if (ev->prev)
+            ev->prev->next = ev->next;
+        free(ev);
+        surface->evacuated = NULL;
+        return;
+    }
+
     unlink_surface (surface);
 
+    if (!surface->cache->all_surfaces) {
+        return;
+    }
+
     if (surface->id != 0					&&
+        surface->host_image                                     &&
 	pixman_image_get_width (surface->host_image) >= 128	&&
 	pixman_image_get_height (surface->host_image) >= 128)
     {
@@ -809,6 +933,16 @@ qxl_surface_flush (qxl_surface_t *surface)
 }
 
 /* access */
+static void
+download_box_no_update (qxl_surface_t *surface, int x1, int y1, int x2, int y2)
+{
+    pixman_image_composite (PIXMAN_OP_SRC,
+                            surface->dev_image,
+                            NULL,
+                            surface->host_image,
+                            x1, y1, 0, 0, x1, y1, x2 - x1, y2 - y1);
+}
+
 static void
 download_box (qxl_surface_t *surface, int x1, int y1, int x2, int y2)
 {
@@ -823,11 +957,7 @@ download_box (qxl_surface_t *surface, int x1, int y1, int x2, int y2)
 
     qxl_update_area(surface->cache->qxl);
 
-    pixman_image_composite (PIXMAN_OP_SRC,
-     			    surface->dev_image,
-			    NULL,
-			    surface->host_image,
-			    x1, y1, 0, 0, x1, y1, x2 - x1, y2 - y1);
+    download_box_no_update(surface, x1, y1, x2, y2);
 }
 
 Bool
@@ -839,9 +969,8 @@ qxl_surface_prepare_access (qxl_surface_t  *surface,
     int n_boxes;
     BoxPtr boxes;
     ScreenPtr pScreen = pixmap->drawable.pScreen;
-    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
     RegionRec new;
-    int stride, height;
 
     if (!pScrn->vtSema)
         return FALSE;
@@ -857,9 +986,6 @@ qxl_surface_prepare_access (qxl_surface_t  *surface,
     n_boxes = REGION_NUM_RECTS (region);
     boxes = REGION_RECTS (region);
 
-    stride = pixman_image_get_stride (surface->dev_image);
-    height = pixman_image_get_height (surface->dev_image);
-
     if (n_boxes < 25)
     {
 	while (n_boxes--)
@@ -1015,7 +1141,7 @@ qxl_surface_cache_evacuate_all (surface_cache_t *cache)
     {
 	if (cache->cached_surfaces[i])
 	{
-	    send_destroy (cache->cached_surfaces[i]);
+            surface_destroy (cache->cached_surfaces[i]);
 	    cache->cached_surfaces[i] = NULL;
 	}
     }
@@ -1044,16 +1170,17 @@ qxl_surface_cache_evacuate_all (surface_cache_t *cache)
 	unlink_surface (s);
 	
 	evacuated->next = evacuated_surfaces;
+        if (evacuated_surfaces)
+            evacuated_surfaces->prev = evacuated;
 	evacuated_surfaces = evacuated;
+        s->evacuated = evacuated;
 
 	s = next;
     }
 
-    free (cache->all_surfaces);
-    cache->all_surfaces = NULL;
     cache->live_surfaces = NULL;
     cache->free_surfaces = NULL;
-    
+
     return evacuated_surfaces;
 }
 
@@ -1209,7 +1336,7 @@ qxl_surface_copy (qxl_surface_t *dest,
     }
     else
     {
-	struct QXLImage *image = qxl_allocnf (qxl, sizeof *image);
+	struct QXLImage *image = qxl_allocnf (qxl, sizeof *image, "surface image struct");
 
 	dest->u.copy_src->ref_count++;
 
@@ -1252,6 +1379,189 @@ qxl_surface_copy (qxl_surface_t *dest,
     push_drawable (qxl, drawable);
 }
 
+/* composite */
+Bool
+qxl_surface_prepare_composite (int op,
+			       PicturePtr	src_picture,
+			       PicturePtr	mask_picture,
+			       PicturePtr	dest_picture,
+			       qxl_surface_t *	src,
+			       qxl_surface_t *	mask,
+			       qxl_surface_t *	dest)
+{
+    dest->u.composite.op = op;
+    dest->u.composite.src_picture = src_picture;
+    dest->u.composite.mask_picture = mask_picture;
+    dest->u.composite.dest_picture = dest_picture;
+    dest->u.composite.src = src;
+    dest->u.composite.mask = mask;
+    dest->u.composite.dest = dest;
+    
+    return TRUE;
+}
+
+static QXLImage *
+image_from_picture (qxl_screen_t *qxl,
+		    PicturePtr picture,
+		    qxl_surface_t *surface,
+		    int *force_opaque)
+{
+    struct QXLImage *image = qxl_allocnf (qxl, sizeof *image, "image struct for picture");
+
+    image->descriptor.id = 0;
+    image->descriptor.type = SPICE_IMAGE_TYPE_SURFACE;
+    image->descriptor.width = 0;
+    image->descriptor.height = 0;
+    image->surface_image.surface_id = surface->id;
+
+    if (picture->format == PICT_x8r8g8b8)
+	*force_opaque = TRUE;
+    else
+	*force_opaque = FALSE;
+    
+    return image;
+}
+
+static QXLTransform *
+get_transform (qxl_screen_t *qxl, PictTransform *transform)
+{
+    if (transform)
+    {
+	QXLTransform *qxform = qxl_allocnf (qxl, sizeof (QXLTransform), "transform");
+
+	qxform->t00 = transform->matrix[0][0];
+	qxform->t01 = transform->matrix[0][1];
+	qxform->t02 = transform->matrix[0][2];
+	qxform->t10 = transform->matrix[1][0];
+	qxform->t11 = transform->matrix[1][1];
+	qxform->t12 = transform->matrix[1][2];
+
+	return qxform;
+    }
+    else
+    {
+	return NULL;
+    }
+}
+
+static QXLRect
+full_rect (qxl_surface_t *surface)
+{
+    QXLRect r;
+    int w = pixman_image_get_width (surface->host_image);
+    int h = pixman_image_get_height (surface->host_image);
+	    
+    r.left = r.top = 0;
+    r.right = w;
+    r.bottom = h;
+
+    return r;
+}
+
+void
+qxl_surface_composite (qxl_surface_t *dest,
+		       int src_x, int src_y,
+		       int mask_x, int mask_y,
+		       int dest_x, int dest_y,
+		       int width, int height)
+{
+    qxl_screen_t *qxl = dest->cache->qxl;
+    PicturePtr src = dest->u.composite.src_picture;
+    qxl_surface_t *qsrc = dest->u.composite.src;
+    PicturePtr mask = dest->u.composite.mask_picture;
+    qxl_surface_t *qmask = dest->u.composite.mask;
+    int op = dest->u.composite.op;
+    struct QXLDrawable *drawable;
+    QXLComposite *composite;
+    QXLRect rect;
+    QXLImage *img;
+    QXLTransform *trans;
+    int n_deps = 0;
+    int force_opaque;
+
+#if 0
+    ErrorF ("QXL Composite: src:       %x (%d %d) id: %d; \n"
+	    "               mask:      id: %d\n"
+	    "               dest:      %x %d %d %d %d (id: %d)\n",
+	    dest->u.composite.src_picture->format,
+	    dest->u.composite.src_picture->pDrawable->width,
+	    dest->u.composite.src_picture->pDrawable->height,
+	    dest->u.composite.src->id,
+	    dest->u.composite.mask? dest->u.composite.mask->id : -1,
+	    dest->u.composite.dest_picture->format,
+	    dest_x, dest_y, width, height,
+	    dest->id
+	);
+#endif
+
+    rect.left = dest_x;
+    rect.right = dest_x + width;
+    rect.top = dest_y;
+    rect.bottom = dest_y + height;
+    
+    drawable = make_drawable (qxl, dest->id, QXL_DRAW_COMPOSITE, &rect);
+
+    composite = &drawable->u.composite;
+
+    composite->flags = 0;
+
+    if (dest->u.composite.dest_picture->format == PICT_x8r8g8b8)
+	composite->flags |= SPICE_COMPOSITE_DEST_OPAQUE;
+    
+    composite->flags |= (op & 0xff);
+
+    img = image_from_picture (qxl, src, qsrc, &force_opaque);
+    if (force_opaque)
+	composite->flags |= SPICE_COMPOSITE_SOURCE_OPAQUE;
+    composite->src = physical_address (qxl, img, qxl->main_mem_slot);
+    composite->flags |= (src->filter << 8);
+    composite->flags |= (src->repeat << 14);
+    trans = get_transform (qxl, src->transform);
+    composite->src_transform = trans?
+	physical_address (qxl, trans, qxl->main_mem_slot) : 0x00000000;
+
+    drawable->surfaces_dest[n_deps] = qsrc->id;
+    drawable->surfaces_rects[n_deps] = full_rect (qsrc);
+
+    n_deps++;
+    
+    if (mask)
+    {
+	img = image_from_picture (qxl, mask, qmask, &force_opaque);
+	if (force_opaque)
+	    composite->flags |= SPICE_COMPOSITE_MASK_OPAQUE;
+	composite->mask = physical_address (qxl, img, qxl->main_mem_slot);
+	composite->flags |= (mask->filter << 11);
+	composite->flags |= (mask->repeat << 16);
+	composite->flags |= (mask->componentAlpha << 18);
+
+	drawable->surfaces_dest[n_deps] = qmask->id;
+	drawable->surfaces_rects[n_deps] = full_rect (qmask);
+	n_deps++;
+	
+	trans = get_transform (qxl, src->transform);
+	composite->mask_transform = trans?
+	    physical_address (qxl, trans, qxl->main_mem_slot) : 0x00000000;
+    }
+    else
+    {
+	composite->mask = 0x00000000;
+	composite->mask_transform = 0x00000000;
+    }
+
+    drawable->surfaces_dest[n_deps] = dest->id;
+    drawable->surfaces_rects[n_deps] = full_rect (dest);
+    
+    composite->src_origin.x = src_x;
+    composite->src_origin.y = src_y;
+    composite->mask_origin.x = mask_x;
+    composite->mask_origin.y = mask_y;
+
+    drawable->effect = QXL_EFFECT_BLEND;
+    
+    push_drawable (qxl, drawable);
+}
+
 Bool
 qxl_surface_put_image (qxl_surface_t *dest,
 		       int x, int y, int width, int height,
